@@ -45,6 +45,7 @@
 #include "partman.h"
 
 #define PARTMAN_SAFETY_MB 32 // 1MB at start + Compensate for rounding errors.
+#define PARTMAN_MAX_PARTS 128 // Maximum number of partitions Linux supports.
 
 PartMan::PartMan(MProcess &mproc, Ui::MeInstall &ui, QWidget *parent)
     : QAbstractItemModel(parent), proc(mproc), root(DeviceItem::Unknown), gui(ui), master(parent)
@@ -80,6 +81,12 @@ void PartMan::setup()
         gui.boxCryptoPass->hide();
         gui.treePartitions->setColumnHidden(Encrypt, true);
     }
+
+    // UUID of the device that the live system is booted from.
+    if (QFile::exists("/live/config/initrd.out")) {
+        QSettings livecfg("/live/config/initrd.out", QSettings::NativeFormat);
+        bootUUID = livecfg.value("BOOT_UUID").toString();
+    }
 }
 
 void PartMan::scan(DeviceItem *drvstart)
@@ -96,12 +103,6 @@ void PartMan::scan(DeviceItem *drvstart)
         root.clear();
     }
 
-    QString bootUUID;
-    if (QFile::exists("/live/config/initrd.out")) {
-        QSettings livecfg("/live/config/initrd.out", QSettings::NativeFormat);
-        bootUUID = livecfg.value("BOOT_UUID").toString();
-    }
-
     for (const QJsonValue &jsonDrive : jsonBD) {
         const QString &jdName = jsonDrive["name"].toString();
         if (jsonDrive["type"] != "disk") continue;
@@ -135,16 +136,16 @@ void PartMan::scan(DeviceItem *drvstart)
             partit->flags.curESP = (partTypeName=="EFI System");
             if (partit->flags.curESP) partit->usefor = "ESP";
             if (!partit->flags.nasty) partit->flags.nasty = partTypeName.startsWith("Microsoft LDM");
-            partit->flags.start = (!bootUUID.isEmpty() && jsonPart["uuid"]==bootUUID);
+            partit->flags.bootRoot = (!bootUUID.isEmpty() && jsonPart["uuid"]==bootUUID);
             partit->curFormat = jsonPart["fstype"].toString();
             // Propagate the boot and nasty flags up to the drive.
-            if (partit->flags.start) drvit->flags.start = true;
+            if (partit->flags.bootRoot) drvit->flags.bootRoot = true;
             if (partit->flags.nasty) drvit->flags.nasty = true;
             notifyChange(partit);
         }
         for (int ixPart = drvit->childCount() - 1; ixPart >= 0; --ixPart) {
             // Propagate the boot flag across the entire drive.
-            if (drvit->flags.start) drvit->child(ixPart)->flags.start = true;
+            if (drvit->flags.bootRoot) drvit->child(ixPart)->flags.bootRoot = true;
         }
         notifyChange(drvit);
     }
@@ -174,8 +175,7 @@ void PartMan::scanVirtualDevices(bool rescan)
         }
     }
     const QString &bdRaw = proc.execOut("lsblk -T -bJo"
-        " TYPE,NAME,UUID,SIZE,PHY-SEC,FSTYPE,LABEL"
-        " /dev/mapper/* 2>/dev/null", true);
+        " TYPE,NAME,UUID,SIZE,PHY-SEC,FSTYPE,LABEL /dev/mapper/* 2>/dev/null", true);
     const QJsonObject &jsonObjBD = QJsonDocument::fromJson(bdRaw.toUtf8()).object();
     const QJsonArray &jsonBD = jsonObjBD["blockdevices"].toArray();
     if (jsonBD.empty()) {
@@ -189,34 +189,37 @@ void PartMan::scanVirtualDevices(bool rescan)
     }
     assert(vdlit != nullptr);
     for (const QJsonValue &jsonDev : jsonBD) {
-        const QString &bdname = jsonDev["name"].toString();
-        const long long bdsize = jsonDev["size"].toVariant().toLongLong();
-        const QString &bdlabel = jsonDev["label"].toString();
-        const bool crypto = jsonDev["type"].toString() == "crypt";
+        const QString &name = jsonDev["name"].toString();
+        const long long size = jsonDev["size"].toVariant().toLongLong();
+        const int physec = jsonDev["phy-sec"].toInt();
+        const QString &label = jsonDev["label"].toString();
+        const bool crypto = (jsonDev["type"].toString() == "crypt");
         // Check if the device is already in the list.
-        DeviceItem *devit = listed.value(bdname);
+        DeviceItem *devit = listed.value(name);
         if (devit) {
-            if (bdsize != devit->size || bdlabel != devit->curLabel
+            if (size != devit->size || physec != devit->physec || label != devit->curLabel
                 || crypto != devit->flags.cryptoV) {
                 // List entry is different to the device, so refresh it.
                 delete devit;
                 devit = nullptr;
             }
-            listed.remove(bdname);
+            listed.remove(name);
         }
         // Create a new list entry if needed.
         if (!devit) {
             devit = new DeviceItem(DeviceItem::VirtualBD, vdlit);
-            devit->device = devit->devMapper = bdname;
-            devit->size = bdsize;
-            devit->curLabel = bdlabel;
+            devit->device = devit->devMapper = name;
+            devit->size = size;
+            devit->physec = physec;
+            devit->flags.bootRoot = (!bootUUID.isEmpty() && jsonDev["uuid"]==bootUUID);
+            devit->curLabel = label;
             devit->curFormat = jsonDev["fstype"].toString();
             devit->flags.cryptoV = crypto;
             notifyChange(devit);
         }
-        DeviceItem *orit = findOrigin(bdname);
+        DeviceItem *orit = findOrigin(name);
         if (orit) {
-            orit->devMapper = bdname;
+            orit->devMapper = name;
             notifyChange(orit);
         }
     }
@@ -245,6 +248,7 @@ bool PartMan::manageConfig(MSettings &config, bool save)
             drvPreserve = false;
             drvit->clear();
             partCount = config.value(configNewLayout).toInt();
+            if (partCount > PARTMAN_MAX_PARTS) return false;
             bootdev = config.value("Boot");
         }
         // Partition configuration.
@@ -270,6 +274,7 @@ bool PartMan::manageConfig(MSettings &config, bool save)
                 if (!partit->format.isEmpty()) config.setValue("Format", partit->format);
                 if (!partit->label.isEmpty()) config.setValue("Label", partit->label);
                 if (!partit->options.isEmpty()) config.setValue("Options", partit->options);
+                config.setValue("CheckBadBlocks", partit->chkbadblk);
                 config.setValue("Dump", partit->dump);
                 config.setValue("Pass", partit->pass);
             } else {
@@ -281,6 +286,7 @@ bool PartMan::manageConfig(MSettings &config, bool save)
                 }
                 partit->usefor = config.value("UseFor").toString();
                 partit->format = config.value("Format").toString();
+                partit->chkbadblk = config.value("CheckBadBlocks").toBool();
                 partit->encrypt = config.value("Encrypt").toBool();
                 partit->label = config.value("Label").toString();
                 partit->options = config.value("Options").toString();
@@ -336,13 +342,9 @@ void PartMan::resizeColumnsToFit()
 void PartMan::treeItemChange()
 {
     // Encryption and bad blocks controls
-    bool canCheck = false;
     bool cryptoAny = false;
     for (DeviceItemIterator it(*this); DeviceItem *item = *it; it.next()) {
         if (item->type != DeviceItem::Partition) continue;
-        if (item->format != "PRESERVE" && !item->usefor.isEmpty()) {
-            if (item->format.startsWith("ext") || item->format == "jfs") canCheck = true;
-        }
         if (item->canEncrypt() && item->encrypt) cryptoAny = true;
     }
     if (gui.boxCryptoPass->isEnabled() != cryptoAny) {
@@ -350,7 +352,6 @@ void PartMan::treeItemChange()
         gui.pushNext->setDisabled(cryptoAny);
         gui.boxCryptoPass->setEnabled(cryptoAny);
     }
-    gui.checkBadBlocks->setEnabled(canCheck);
     treeSelChange();
 }
 
@@ -373,11 +374,8 @@ void PartMan::treeSelChange()
         if (!drvit) drvit = twit;
         if (!islocked && isold && isdrive) gui.pushPartAdd->setEnabled(false);
         else if (!isold) {
-            long long maxMB = (drvit->size / 1048576) - PARTMAN_SAFETY_MB;
-            for (int ixi = drvit->childCount() - 1; ixi >= 0; --ixi) {
-                maxMB -= drvit->child(ixi)->size / 1048576;
-            }
-            gui.pushPartAdd->setEnabled(maxMB > 0);
+            gui.pushPartAdd->setEnabled(drvit->childCount() < PARTMAN_MAX_PARTS
+                && twit->driveFreeSpace(true) > 1048576);
         }
     } else {
         gui.pushPartClear->setEnabled(false);
@@ -497,13 +495,14 @@ void PartMan::partAddClick(bool)
     DeviceItem *twit = (indexes.size() > 0) ? item(indexes.at(0)) : nullptr;
     if (!twit) return;
     DeviceItem *drive = twit->parent();
-    if (!drive || drive->type == DeviceItem::Unknown) drive = twit;
+    if (!drive) drive = twit;
 
     DeviceItem *part = new DeviceItem(DeviceItem::Partition, drive, twit);
     drive->labelParts();
     drive->flags.oldLayout = false;
     notifyChange(drive);
-    gui.treePartitions->selectionModel()->select(index(part), QItemSelectionModel::ClearAndSelect);
+    gui.treePartitions->selectionModel()->select(index(part),
+        QItemSelectionModel::ClearAndSelect | QItemSelectionModel::Rows);
 }
 
 void PartMan::partRemoveClick(bool)
@@ -631,7 +630,8 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
                 return false;
             }
             // Check for duplicate subvolume label entries.
-            DeviceItem *pit = item->parent();
+            DeviceItem *pit = item->parentItem;
+            assert(pit != nullptr);
             const int count = pit->childCount();
             const int index = pit->indexOfChild(item);
             for (int ixi = 0; ixi < count; ++ixi) {
@@ -663,8 +663,9 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
         } else if(!mount.isEmpty() && mount != "FORMAT" && mount != "ESP"){
             mounts.insert(mount, *it);
         }
-        if (item->type != DeviceItem::VirtualBD && item->encrypt) {
-            if (mount == "/") item->devMapper = "root.fsm";
+        if (item->type != DeviceItem::VirtualBD) {
+            if (!item->encrypt) item->devMapper.clear();
+            else if (mount == "/") item->devMapper = "root.fsm";
             else if (mount.startsWith("SWAP")) item->devMapper = mount.toLower();
             else item->devMapper = QString::number(++mapnum) + mount.replace('/','.') + ".fsm";
         }
@@ -676,7 +677,13 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
     }
 
     DeviceItem *rootitem = mounts.value("/");
-    if (rootitem) {
+    if (!rootitem || rootitem->size < rootSpaceNeeded) {
+        const QString &tmin = QLocale::system().formattedDataSize(rootSpaceNeeded + 1048575,
+            1, QLocale::DataSizeTraditionalFormat);
+        QMessageBox::critical(master, master->windowTitle(),
+            tr("A root partition of at least %1 is required.").arg(tmin));
+        return false;
+    } else {
         if (!rootitem->willFormat() && mounts.contains("/home")) {
             const QString errmsg = tr("Cannot preserve /home inside root (/) if"
                 " a separate /home partition is also mounted.");
@@ -684,12 +691,6 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
             return false;
         }
         if (rootitem->encrypt) encryptRoot = true;
-    } else {
-        const QString &tmin = QLocale::system().formattedDataSize(rootSpaceNeeded + 1048575,
-            1, QLocale::DataSizeTraditionalFormat);
-        QMessageBox::critical(master, master->windowTitle(), tr("You must choose a root partition.\n"
-            "The root partition must be at least %1.").arg(tmin));
-        return false;
     }
     if (encryptRoot && !mounts.contains("/boot")) {
         QMessageBox::critical(master, master->windowTitle(),
@@ -712,17 +713,19 @@ bool PartMan::composeValidate(bool automatic, const QString &project)
             bool hasBiosGrub = false;
             for (int ixdev = 0; ixdev < partCount; ++ixdev) {
                 DeviceItem *partit = drvit->child(ixdev);
-                const QString &dev = partit->shownDevice();
                 const QString &use = partit->realUseFor();
-                if (use.isEmpty()) continue;
-                else if (use == "BIOS-GRUB") hasBiosGrub = true;
-                if (partit->willFormat()) details += tr("Format %1").arg(dev);
-                else if (use == "/") {
-                    details += tr("Delete the data on %1 except for /home").arg(dev);
+                QString actmsg;
+                if (drvit->flags.oldLayout) {
+                    if (use.isEmpty()) continue;
+                    else if (partit->willFormat()) actmsg = tr("Format %1 to use for %2");
+                    else if (use != "/") actmsg = tr("Reuse (no reformat) %1 as %2");
+                    else actmsg = tr("Delete the data on %1 except for /home, to use for %2");
                 } else {
-                    details += tr("Reuse (no reformat) %1 as %2").arg(dev, partit->shownUseFor());
+                    if (use.isEmpty()) actmsg = tr("Create %1 without formatting");
+                    else actmsg = tr("Create %1, format to use for %2");
                 }
-                details += '\n';
+                details += actmsg.arg(partit->shownDevice(), partit->shownUseFor()) + '\n';
+                if (use == "BIOS-GRUB") hasBiosGrub = true;
             }
             // Potentially unbootable GPT when on a BIOS-based PC.
             const bool hasBoot = (drvit->active != nullptr);
@@ -985,8 +988,6 @@ bool PartMan::formatPartitions()
         ? gui.textCryptoPass : gui.textCryptoPassCust)->text().toUtf8();
 
     // Format partitions.
-    const bool badblocks = gui.checkBadBlocks->isChecked();
-
     for (DeviceItemIterator it(*this); DeviceItem *twit = *it; it.next()) {
         if (twit->type != DeviceItem::Partition || !twit->willFormat()) continue;
         const QString &dev = twit->mappedDevice(true);
@@ -1016,7 +1017,7 @@ bool PartMan::formatPartitions()
             if (!twit->label.isEmpty()) cmd.append(" -L \"" + twit->label + '"');
             if (!proc.exec(cmd, true)) return false;
         } else {
-            if (!formatLinuxPartition(dev, twit->format, badblocks, twit->label)) return false;
+            if (!formatLinuxPartition(dev, twit->format, twit->chkbadblk, twit->label)) return false;
         }
     }
     // Prepare subvolumes on all that (are to) contain them.
@@ -1331,54 +1332,70 @@ QVariant PartMan::data(const QModelIndex &index, int role) const
 {
     DeviceItem *item = static_cast<DeviceItem*>(index.internalPointer());
     const bool isDriveOrVD = (item->type == DeviceItem::Drive || item->type == DeviceItem::VirtualDevices);
-    if (role == Qt::DisplayRole) {
-        switch (index.column()) {
-            case Device:
-                if (item->type == DeviceItem::Subvolume) return "----";
-                else return item->device;
-                break;
-            case Size:
-                if (item->type == DeviceItem::Subvolume) return "----";
-                else {
-                    return QLocale::system().formattedDataSize(item->size,
-                        1, QLocale::DataSizeTraditionalFormat);
-                }
-                break;
-            case UseFor: return item->usefor; break;
-            case Label:
-                if (item->type == DeviceItem::Drive) return item->model;
-                else if (index.flags() & Qt::ItemIsEditable) return item->label;
-                else return item->curLabel;
-                break;
-            case Format:
-                if (item->usefor.isEmpty()) return item->curFormat;
-                else return item->shownFormat(item->format);
-                break;
-            case Options:
-                if (item->canMount() || item->realUseFor() == "SWAP") return item->options;
-                else if (!isDriveOrVD) return "--------";
-                break;
-            case Pass:
-                if (item->canMount()) return item->pass;
-                else if (!isDriveOrVD) return "--";
-                break;
-        }
-    } else if (role == Qt::EditRole) {
-        switch (index.column()) {
-            case Device: return item->device; break;
-            case Size: return item->size; break;
-            case UseFor: return item->usefor; break;
-            case Label: return item->label; break;
-            case Format: return item->format; break;
-            case Options: return item->options; break;
-            case Pass: return item->pass; break;
+    if (role == Qt::EditRole) {
+        switch (index.column())
+        {
+        case Device: return item->device; break;
+        case Size: return item->size; break;
+        case UseFor: return item->usefor; break;
+        case Label: return item->label; break;
+        case Format: return item->format; break;
+        case Options: return item->options; break;
+        case Pass: return item->pass; break;
         }
     } else if (role == Qt::CheckStateRole && !isDriveOrVD
         && index.flags() & Qt::ItemIsUserCheckable) {
-        switch (index.column()) {
-            case Encrypt: return item->encrypt ? Qt::Checked : Qt::Unchecked; break;
-            case Dump: return item->dump ? Qt::Checked : Qt::Unchecked; break;
+        switch (index.column())
+        {
+        case Encrypt: return item->encrypt ? Qt::Checked : Qt::Unchecked; break;
+        case Check: return item->chkbadblk ? Qt::Checked : Qt::Unchecked; break;
+        case Dump: return item->dump ? Qt::Checked : Qt::Unchecked; break;
         }
+    } else if (role == Qt::DisplayRole) {
+        switch (index.column()) {
+        case Device:
+            if (item->type == DeviceItem::Subvolume) return "----";
+            else return item->device;
+            break;
+        case Size:
+            if (item->type == DeviceItem::Subvolume) return "----";
+            else {
+                return QLocale::system().formattedDataSize(item->size,
+                    1, QLocale::DataSizeTraditionalFormat);
+            }
+            break;
+        case UseFor: return item->usefor; break;
+        case Label:
+            if (index.flags() & Qt::ItemIsEditable) return item->label;
+            else return item->curLabel;
+            break;
+        case Format:
+            if (item->usefor.isEmpty()) return item->curFormat;
+            else return item->shownFormat(item->format);
+            break;
+        case Options:
+            if (item->canMount() || item->realUseFor() == "SWAP") return item->options;
+            else if (!isDriveOrVD) return "--------";
+            break;
+        case Pass:
+            if (item->canMount()) return item->pass;
+            else if (!isDriveOrVD) return "--";
+            break;
+        }
+    } else if (role == Qt::ToolTipRole) {
+        switch (index.column()) {
+        case Device: if (!item->model.isEmpty()) return tr("Model: %1").arg(item->model); break;
+        case Size:
+            if (item->type == DeviceItem::Subvolume) return "----";
+            else if (item->type == DeviceItem::Drive) {
+                long long fs = item->driveFreeSpace();
+                if (fs <= 0) fs = 0;
+                return tr("Free space: %1").arg(QLocale::system().formattedDataSize(fs,
+                    1, QLocale::DataSizeTraditionalFormat));
+            }
+            break;
+        }
+        return item->shownDevice();
     } else if (role == Qt::DecorationRole && index.column() == Device) {
         if (item->type == DeviceItem::Drive && !item->flags.oldLayout) {
             return QIcon(":/appointment-soon");
@@ -1399,9 +1416,11 @@ bool PartMan::setData(const QModelIndex &index, const QVariant &value, int role)
     if (role == Qt::CheckStateRole) {
         DeviceItem *item = static_cast<DeviceItem *>(index.internalPointer());
         changeBegin(item);
-        switch (index.column()) {
-            case Encrypt: item->encrypt = (value == Qt::Checked); break;
-            case Dump: item->dump = (value == Qt::Checked); break;
+        switch (index.column())
+        {
+        case Encrypt: item->encrypt = (value == Qt::Checked); break;
+        case Check: item->chkbadblk = (value == Qt::Checked); break;
+        case Dump: item->dump = (value == Qt::Checked); break;
         }
     }
     if(!changeEnd()) emit dataChanged(index, index);
@@ -1413,38 +1432,44 @@ Qt::ItemFlags PartMan::flags(const QModelIndex &index) const
     if (item->type == DeviceItem::VirtualDevices) return Qt::ItemIsEnabled;
     Qt::ItemFlags flagsOut({Qt::ItemIsSelectable, Qt::ItemIsEnabled});
     if (item->mapCount) return flagsOut;
-    switch (index.column()) {
-        case Device: break;
-        case Size:
-            if (item->type == DeviceItem::Partition && !item->flags.oldLayout) {
-                flagsOut |= Qt::ItemIsEditable;
-            }
-            break;
-        case UseFor:
-            if (item->type != DeviceItem::Drive) flagsOut |= Qt::ItemIsEditable;
-            break;
-        case Label:
-            if (item->type == DeviceItem::Subvolume) {
-                if (item->format != "PRESERVE") flagsOut |= Qt::ItemIsEditable;
-            } else {
-                if (!item->usefor.isEmpty()) flagsOut |= Qt::ItemIsEditable;
-            }
-            break;
-        case Encrypt:
-            if (item->canEncrypt()) flagsOut |= Qt::ItemIsUserCheckable;
-            break;
-        case Format:
+    switch (index.column())
+    {
+    case Device: break;
+    case Size:
+        if (item->type == DeviceItem::Partition && !item->flags.oldLayout) {
+            flagsOut |= Qt::ItemIsEditable;
+        }
+        break;
+    case UseFor:
+        if (item->allowedUsesFor().count() > 1) flagsOut |= Qt::ItemIsEditable;
+        break;
+    case Label:
+        if (item->type == DeviceItem::Subvolume) {
+            if (item->format != "PRESERVE") flagsOut |= Qt::ItemIsEditable;
+        } else {
             if (!item->usefor.isEmpty()) flagsOut |= Qt::ItemIsEditable;
-            break;
-        case Options:
-            if (item->canMount() || item->realUseFor() == "SWAP") flagsOut |= Qt::ItemIsEditable;
-            break;
-        case Dump:
-            if (item->canMount()) flagsOut |= Qt::ItemIsUserCheckable;
-            break;
-        case Pass:
-            if (item->canMount()) flagsOut |= Qt::ItemIsEditable;
-            break;
+        }
+        break;
+    case Encrypt:
+        if (item->canEncrypt()) flagsOut |= Qt::ItemIsUserCheckable;
+        break;
+    case Format:
+        if (item->allowedFormats().count() > 1) flagsOut |= Qt::ItemIsEditable;
+        break;
+    case Check:
+        if (item->format.startsWith("ext") || item->format == "jfs") {
+            flagsOut |= Qt::ItemIsUserCheckable;
+        }
+        break;
+    case Options:
+        if (item->canMount() || item->realUseFor() == "SWAP") flagsOut |= Qt::ItemIsEditable;
+        break;
+    case Dump:
+        if (item->canMount()) flagsOut |= Qt::ItemIsUserCheckable;
+        break;
+    case Pass:
+        if (item->canMount()) flagsOut |= Qt::ItemIsEditable;
+        break;
     }
     return flagsOut;
 }
@@ -1452,18 +1477,20 @@ QVariant PartMan::headerData(int section, Qt::Orientation orientation, int role)
 {
     assert(orientation == Qt::Horizontal);
     if (role == Qt::DisplayRole) {
-        switch (section) {
-            case Device: return tr("Device"); break;
-            case Size: return tr("Size"); break;
-            case UseFor: return tr("Use For"); break;
-            case Label: return tr("Label"); break;
-            case Encrypt: return tr("Encrypt"); break;
-            case Format: return tr("Format"); break;
-            case Options: return tr("Options"); break;
-            case Dump: return tr("Dump"); break;
-            case Pass: return tr("Pass"); break;
+        switch (section)
+        {
+        case Device: return tr("Device"); break;
+        case Size: return tr("Size"); break;
+        case UseFor: return tr("Use For"); break;
+        case Label: return tr("Label"); break;
+        case Encrypt: return tr("Encrypt"); break;
+        case Format: return tr("Format"); break;
+        case Check: return tr("Check"); break;
+        case Options: return tr("Options"); break;
+        case Dump: return tr("Dump"); break;
+        case Pass: return tr("Pass"); break;
         }
-    } else if (role == Qt::FontRole && section == Encrypt) {
+    } else if (role == Qt::FontRole && (section == Encrypt || section == Check || section == Dump)) {
         QFont smallFont;
         smallFont.setPointSizeF(smallFont.pointSizeF() * 0.6);
         return smallFont;
@@ -1487,7 +1514,7 @@ QModelIndex PartMan::parent(const QModelIndex &index) const
 {
     if (!index.isValid()) return QModelIndex();
     DeviceItem *cit = static_cast<DeviceItem *>(index.internalPointer());
-    DeviceItem *pit = cit->parent();
+    DeviceItem *pit = cit->parentItem;
     if (!pit || pit == &root) return QModelIndex();
     return createIndex(pit->row(), 0, pit);
 }
@@ -1563,7 +1590,7 @@ int PartMan::changeEnd(bool notify)
 }
 void PartMan::notifyChange(class DeviceItem *item, int first, int last)
 {
-    DeviceItem *p = item->parent();
+    DeviceItem *const p = item->parentItem;
     if (first < 0) first = 0;
     if (last < 0) last = p ? p->childCount() - 1 : 0;
     const int row = p ? p->indexOfChild(item) : 0;
@@ -1575,6 +1602,7 @@ void PartMan::notifyChange(class DeviceItem *item, int first, int last)
 DeviceItem::DeviceItem(enum DeviceType type, DeviceItem *parent, DeviceItem *preceding)
     : parentItem(parent), type(type)
 {
+    if (type == Partition) size = 1048576;
     if (parent) {
         partman = parent->partman;
         const int i = preceding ? (parentItem->children.indexOf(preceding) + 1) : parentItem->childCount();
@@ -1624,6 +1652,7 @@ inline int DeviceItem::row() const
 }
 inline DeviceItem *DeviceItem::parent() const
 {
+    if (parentItem && !parentItem->parentItem) return nullptr; // Invisible root
     return parentItem;
 }
 inline DeviceItem *DeviceItem::child(int row) const
@@ -1678,7 +1707,7 @@ void DeviceItem::setActive(bool boot)
 }
 inline bool DeviceItem::isActive() const
 {
-    if (parentItem) return false;
+    if (!parentItem) return false;
     return (parentItem->active == this);
 }
 bool DeviceItem::isLocked() const
@@ -1702,7 +1731,7 @@ bool DeviceItem::canEncrypt() const
 QString DeviceItem::mappedDevice(const bool full) const
 {
     const DeviceItem *twit = this;
-    if (twit->type == Subvolume) twit = twit->parent();
+    if (twit->type == Subvolume) twit = twit->parentItem;
     if (twit->type == Partition) {
         const QVariant &d = twit->devMapper;
         if (!d.isNull()) {
@@ -1725,6 +1754,7 @@ QString DeviceItem::shownDevice() const
 }
 QStringList DeviceItem::allowedUsesFor(bool real) const
 {
+    if (!isVolume()) return QStringList();
     QStringList list;
     if (!partman || size >= partman->rootSpaceNeeded) list << "root";
     if (type == Subvolume) list << "home"; // swap requires Linux 5.0 or later
@@ -1743,12 +1773,14 @@ QStringList DeviceItem::allowedUsesFor(bool real) const
 }
 QStringList DeviceItem::allowedFormats() const
 {
+    if (!isVolume()) return QStringList();
     QStringList list;
     bool allowPreserve = false, selPreserve = false;
     if (type == Subvolume) list << "CREATE";
     else {
         const QString &use = realUseFor();
-        if (use == "/boot") list << "ext4";
+        if (use.isEmpty()) return QStringList();
+        else if (use == "/boot") list << "ext4";
         else if (use == "BIOS-GRUB") list << "GRUB";
         else if (use == "ESP") {
             list << "FAT32" << "FAT16" << "FAT12";
@@ -1786,6 +1818,17 @@ bool DeviceItem::canMount() const
     const QString &use = realUseFor();
     return !(use.isEmpty() || use == "FORMAT" || use == "ESP" || use == "SWAP");
 }
+long long DeviceItem::driveFreeSpace(bool inclusive) const
+{
+    const DeviceItem *drvit = parent();
+    if (!drvit) drvit = this;
+    long long free = drvit->size - (PARTMAN_SAFETY_MB * 1048576);
+    for (int ixi = drvit->children.count() - 1; ixi >= 0; --ixi) {
+        DeviceItem *partit = drvit->children.at(ixi);
+        if (inclusive || partit != this) free -= partit->size;
+    }
+    return free;
+}
 /* Convenience */
 DeviceItem *DeviceItem::addPart(int defaultMB, const QString &defaultUse, bool crypto)
 {
@@ -1990,14 +2033,14 @@ void DeviceItemIterator::next()
         ixPos = 0;
         pos = pos->child(0);
     } else {
-        DeviceItem *parent = pos->parent();
+        DeviceItem *parent = pos->parentItem;
         if (!parent) {
             pos = nullptr;
             return;
         }
         DeviceItem *chnext = parent->child(ixPos+1);
         while (!chnext && parent) {
-            parent = parent->parent();
+            parent = parent->parentItem;
             if (!parent) break;
             ixPos = ixParents.pop();
             chnext = parent->child(ixPos+1);
@@ -2026,73 +2069,91 @@ void DeviceItemDelegate::paint(QPainter *painter, const QStyleOptionViewItem &op
 }
 QSize DeviceItemDelegate::sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const
 {
-    const QSize &size = QStyledItemDelegate::sizeHint(option,  index);
-    return QSize(size.width() + 4, option.fontMetrics.height() + 4);
+    DeviceItem *item = static_cast<DeviceItem*>(index.internalPointer());
+    int width = QStyledItemDelegate::sizeHint(option, index).width();
+    // In the case of Format and Use For, the cell should accommodate all options in the list.
+    const int residue = width - option.fontMetrics.boundingRect(index.data(Qt::DisplayRole).toString()).width();
+    switch(index.column()) {
+    case PartMan::Format:
+        for (const QString &fmt : item->allowedFormats()) {
+            const int fw = option.fontMetrics.boundingRect(item->shownFormat(fmt)).width() + residue;
+            if (fw > width) width = fw;
+        }
+        break;
+    case PartMan::UseFor:
+        for (const QString &use : item->allowedUsesFor(false)) {
+            const int uw = option.fontMetrics.boundingRect(use).width() + residue;
+            if (uw > width) width = uw;
+        }
+        break;
+    }
+    return QSize(width + 4, option.fontMetrics.height() + 4);
 }
 
 QWidget *DeviceItemDelegate::createEditor(QWidget *parent, const QStyleOptionViewItem &, const QModelIndex &index) const
 {
     QWidget *widget = nullptr;
-    switch (index.column()) {
-        case PartMan::Size: widget = new QSpinBox(parent); break;
-        case PartMan::UseFor: {
-            QComboBox *combo = new QComboBox(parent);
-            combo->setEditable(true);
-            combo->setInsertPolicy(QComboBox::NoInsert);
-            combo->lineEdit()->setPlaceholderText("----");
-            widget = combo;
-            break;
-        }
-        case PartMan::Format: widget = new QComboBox(parent); break;
-        case PartMan::Pass: widget = new QSpinBox(parent); break;
-        case PartMan::Options: {
+    QComboBox *combo = nullptr;
+    switch (index.column())
+    {
+    case PartMan::Size: widget = new QSpinBox(parent); break;
+    case PartMan::UseFor:
+        widget = combo = new QComboBox(parent);
+        combo->setEditable(true);
+        combo->setInsertPolicy(QComboBox::NoInsert);
+        combo->lineEdit()->setPlaceholderText("----");
+        break;
+    case PartMan::Format: widget = combo = new QComboBox(parent); break;
+    case PartMan::Pass: widget = new QSpinBox(parent); break;
+    case PartMan::Options:
+        {
             QLineEdit *edit = new QLineEdit(parent);
             edit->setProperty("row", QVariant::fromValue<void *>(index.internalPointer()));
             edit->setContextMenuPolicy(Qt::CustomContextMenu);
             connect(edit, &QLineEdit::customContextMenuRequested,
                 this, &DeviceItemDelegate::partOptionsMenu);
             widget = edit;
-            break;
         }
-        default: widget = new QLineEdit(parent);
+        break;
+    default: widget = new QLineEdit(parent);
     }
     assert(widget != nullptr);
     widget->setAutoFillBackground(true);
     widget->setFocusPolicy(Qt::StrongFocus);
+    if (combo) {
+        connect(combo,  QOverload<int>::of(&QComboBox::activated),
+            this, &DeviceItemDelegate::emitCommit);
+    }
     return widget;
 }
 void DeviceItemDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const
 {
     DeviceItem *item = static_cast<DeviceItem*>(index.internalPointer());
-    switch (index.column()) {
-        case PartMan::Size: {
-            DeviceItem *drvit = item->parent();
-            assert(drvit != nullptr);
-            long long max = drvit->size - (PARTMAN_SAFETY_MB*1048576);
-            for (int ixi = drvit->childCount() - 1; ixi >= 0; --ixi) {
-                DeviceItem *partit = drvit->child(ixi);
-                if (partit != item) max -= partit->size;
-            }
+    switch (index.column())
+    {
+    case PartMan::Size:
+        {
             QSpinBox *spin = qobject_cast<QSpinBox *>(editor);
-            spin->setRange(1, static_cast<int>(max / 1048576));
+            spin->setRange(1, static_cast<int>(item->driveFreeSpace() / 1048576));
             spin->setSuffix("MB");
             spin->setStepType(QSpinBox::AdaptiveDecimalStepType);
             spin->setAccelerated(true);
             spin->setValue(item->size / 1048576);
-            break;
         }
-        case PartMan::UseFor: {
+        break;
+    case PartMan::UseFor:
+        {
             QComboBox *combo = qobject_cast<QComboBox *>(editor);
             combo->addItem("");
             combo->addItems(item->allowedUsesFor(false));
             combo->setCurrentText(item->usefor);
-            break;
         }
-        case PartMan::Label: {
-            qobject_cast<QLineEdit *>(editor)->setText(item->label);
-            break;
-        }
-        case PartMan::Format: {
+        break;
+    case PartMan::Label:
+        qobject_cast<QLineEdit *>(editor)->setText(item->label);
+        break;
+    case PartMan::Format:
+        {
             QComboBox *combo = qobject_cast<QComboBox *>(editor);
             const QStringList &formats = item->allowedFormats();
             assert(!formats.isEmpty());
@@ -2106,17 +2167,14 @@ void DeviceItemDelegate::setEditorData(QWidget *editor, const QModelIndex &index
             }
             const int ixfmt = combo->findData(item->format);
             if (ixfmt >= 0) combo->setCurrentIndex(ixfmt);
-            break;
-        }
-        case PartMan::Pass: {
-            QSpinBox *spin = qobject_cast<QSpinBox *>(editor);
-            spin->setValue(item->pass);
-            break;
-        }
-        case PartMan::Options: {
-            qobject_cast<QLineEdit *>(editor)->setText(item->options);
-            break;
         }
+        break;
+    case PartMan::Pass:
+        qobject_cast<QSpinBox *>(editor)->setValue(item->pass);
+        break;
+    case PartMan::Options:
+        qobject_cast<QLineEdit *>(editor)->setText(item->options);
+        break;
     }
 }
 void DeviceItemDelegate::setModelData(QWidget *editor, QAbstractItemModel *model, const QModelIndex &index) const
@@ -2124,33 +2182,38 @@ void DeviceItemDelegate::setModelData(QWidget *editor, QAbstractItemModel *model
     DeviceItem *item = static_cast<DeviceItem*>(index.internalPointer());
     PartMan *partman = qobject_cast<PartMan *>(model);
     partman->changeBegin(item);
-    switch (index.column()) {
-        case PartMan::Size:
-            item->size = qobject_cast<QSpinBox *>(editor)->value();
-            item->size *= 1048576; // Separate step to prevent int overflow.
-            break;
-        case PartMan::UseFor:
-            item->usefor = qobject_cast<QComboBox *>(editor)->currentText().trimmed();
-            break;
-        case PartMan::Label:
-            item->label = qobject_cast<QLineEdit *>(editor)->text().trimmed();
-            break;
-        case PartMan::Format:
-            item->format = qobject_cast<QComboBox *>(editor)->currentData().toString();
-            break;
-        case PartMan::Pass:
-            item->pass = qobject_cast<QSpinBox *>(editor)->value();
-            break;
-        case PartMan::Options:
-            item->options = qobject_cast<QLineEdit *>(editor)->text().trimmed();
-            break;
+    switch (index.column())
+    {
+    case PartMan::Size:
+        item->size = qobject_cast<QSpinBox *>(editor)->value();
+        item->size *= 1048576; // Separate step to prevent int overflow.
+        break;
+    case PartMan::UseFor:
+        item->usefor = qobject_cast<QComboBox *>(editor)->currentText().trimmed();
+        break;
+    case PartMan::Label:
+        item->label = qobject_cast<QLineEdit *>(editor)->text().trimmed();
+        break;
+    case PartMan::Format:
+        item->format = qobject_cast<QComboBox *>(editor)->currentData().toString();
+        break;
+    case PartMan::Pass:
+        item->pass = qobject_cast<QSpinBox *>(editor)->value();
+        break;
+    case PartMan::Options:
+        item->options = qobject_cast<QLineEdit *>(editor)->text().trimmed();
+        break;
     }
     const int changed = partman->changeEnd(false);
     item->autoFill(changed);
     if (changed) partman->notifyChange(item);
 }
+void DeviceItemDelegate::emitCommit()
+{
+    emit commitData(qobject_cast<QWidget *>(sender()));
+}
 
-void DeviceItemDelegate::partOptionsMenu(const QPoint &)
+void DeviceItemDelegate::partOptionsMenu()
 {
     QLineEdit *edit = static_cast<QLineEdit *>(sender());
     if (!edit) return;
